@@ -49,7 +49,7 @@ from util import NotEnoughFunds, PrintError, UserCancelled, profiler
 
 from bitcoin import *
 from version import *
-from keystore import load_keystore
+from keystore import load_keystore, Hardware_KeyStore
 from storage import multisig_type
 
 from transaction import Transaction
@@ -804,6 +804,10 @@ class Abstract_Wallet(PrintError):
         f = self.network.relay_fee if self.network and self.network.relay_fee else RELAY_FEE
         return min(f, MAX_RELAY_FEE)
 
+    def dust_threshold(self):
+        # Change <= dust threshold is added to the tx fee
+        return 182 * 3 * self.relayfee() / 1000
+
     def get_tx_fee(self, tx):
         # this method can be overloaded
         return tx.get_fee()
@@ -844,14 +848,11 @@ class Abstract_Wallet(PrintError):
         else:
             fee_estimator = lambda size: fixed_fee
 
-        # Change <= dust threshold is added to the tx fee
-        dust_threshold = 182 * 3 * self.relayfee() / 1000
-
         # Let the coin chooser select the coins to spend
         max_change = self.max_change_outputs if self.multiple_change else 1
         coin_chooser = coinchooser.get_coin_chooser(config)
         tx = coin_chooser.make_tx(coins, outputs, change_addrs[:max_change],
-                                  fee_estimator, dust_threshold)
+                                  fee_estimator, self.dust_threshold())
 
         # Sort the inputs and outputs deterministically
         tx.BIP_LI01_sort()
@@ -1008,21 +1009,33 @@ class Abstract_Wallet(PrintError):
 
     def bump_fee(self, tx, delta):
         if tx.is_final():
-            raise BaseException("cannot bump fee: transaction is final")
+            raise BaseException(_("Cannot bump fee: transaction is final"))
         inputs = copy.deepcopy(tx.inputs())
         outputs = copy.deepcopy(tx.outputs())
         for txin in inputs:
             txin['signatures'] = [None] * len(txin['signatures'])
             self.add_input_info(txin)
-        for i, o in enumerate(outputs):
+        # use own outputs
+        s = filter(lambda x: self.is_mine(x[1]), outputs)
+        # ... unless there is none
+        if not s:
+            s = outputs
+        # prioritize low value outputs, to get rid of dust
+        s = sorted(s, key=lambda x: x[2])
+        for o in s:
+            i = outputs.index(o)
             otype, address, value = o
-            if self.is_mine(address) and value >= delta:
+            if value - delta >= self.dust_threshold():
                 outputs[i] = otype, address, value - delta
                 break
-        else:
-            raise BaseException("cannot bump fee: could not find a change output")
-        new_tx = Transaction.from_io(inputs, outputs)
-        return new_tx
+            else:
+                del outputs[i]
+                delta -= value
+                if delta > 0:
+                    continue
+        if delta > 0:
+            raise BaseException(_('Cannot bump fee: cound not find suitable outputs'))
+        return Transaction.from_io(inputs, outputs)
 
     def add_input_info(self, txin):
         # Add address for utxo that are in wallet
@@ -1053,15 +1066,11 @@ class Abstract_Wallet(PrintError):
             tx = Transaction(self.network.synchronous_get(request))
         return tx
 
-    def sign_transaction(self, tx, password):
-        if self.is_watching_only():
-            return
-
+    def add_hw_info(self, tx):
         # add previous tx for hw wallets
         for txin in tx.inputs():
             tx_hash = txin['prevout_hash']
             txin['prev_tx'] = self.get_input_tx(tx_hash)
-
         # add output info for hw wallets
         info = {}
         xpubs = self.get_master_public_keys()
@@ -1075,6 +1084,12 @@ class Abstract_Wallet(PrintError):
                 info[addr] = index, sorted_xpubs, self.m if isinstance(self, Multisig_Wallet) else None
         tx.output_info = info
 
+    def sign_transaction(self, tx, password):
+        if self.is_watching_only():
+            return
+        # hardware wallets require extra info
+        if any([(isinstance(k, Hardware_KeyStore) and k.can_sign(tx)) for k in self.get_keystores()]):
+            self.add_hw_info(tx)
         # sign
         for k in self.get_keystores():
             try:
