@@ -34,10 +34,11 @@ from urllib import quote
 
 import electrum
 from electrum import bitcoin
+from electrum import keystore
 from electrum.bitcoin import *
 from electrum.mnemonic import Mnemonic
 from electrum import version
-from electrum.wallet import Multisig_Wallet, BIP32_Wallet
+from electrum.wallet import Multisig_Wallet, Deterministic_Wallet, Wallet
 from electrum.i18n import _
 from electrum.plugins import BasePlugin, run_hook, hook
 from electrum.util import NotEnoughFunds
@@ -73,9 +74,9 @@ class TrustedCoinException(Exception):
         self.status_code = status_code
 
 class TrustedCoinCosignerClient(object):
-    def __init__(self, user_agent=None, base_url='https://api.trustedcoin.com/2/', debug=False):
+    def __init__(self, user_agent=None, base_url='https://api.trustedcoin.com/2/'):
         self.base_url = base_url
-        self.debug = debug
+        self.debug = False
         self.user_agent = user_agent
 
     def send_request(self, method, relative_url, data=None):
@@ -140,13 +141,18 @@ class TrustedCoinCosignerClient(object):
         return self.send_request('post', 'cosigner/%s/auth' % quote(id), payload)
 
     def get(self, id):
-        """
-        Attempt to authenticate for a particular cosigner.
-        :param id: the id of the cosigner
-        :param otp: the one time password
-        """
+        """ Get billing info """
         return self.send_request('get', 'cosigner/%s' % quote(id))
 
+    def get_challenge(self, id):
+        """ Get challenge to reset Google Auth secret """
+        return self.send_request('get', 'cosigner/%s/otp_secret' % quote(id))
+
+    def reset_auth(self, id, challenge, signatures):
+        """ Reset Google Auth secret """
+        payload = {'challenge':challenge, 'signatures':signatures}
+        return self.send_request('post', 'cosigner/%s/otp_secret' % quote(id), payload)
+
     def sign(self, id, transaction, otp):
         """
         Attempt to authenticate for a particular cosigner.
@@ -187,29 +193,16 @@ server = TrustedCoinCosignerClient(user_agent="Electrum/" + version.ELECTRUM_VER
 class Wallet_2fa(Multisig_Wallet):
 
     def __init__(self, storage):
-        BIP32_Wallet.__init__(self, storage)
-        self.wallet_type = '2fa'
-        self.m = 2
-        self.n = 3
+        self.m, self.n = 2, 3
+        Deterministic_Wallet.__init__(self, storage)
         self.is_billing = False
         self.billing_info = None
 
-    def get_action(self):
-        xpub1 = self.master_public_keys.get("x1/")
-        xpub2 = self.master_public_keys.get("x2/")
-        xpub3 = self.master_public_keys.get("x3/")
-        if xpub2 is None and not self.storage.get('use_trustedcoin'):
-            return 'show_disclaimer'
-        if xpub2 is None:
-            return 'create_extended_seed'
-        if xpub3 is None:
-            return 'create_remote_key'
-
-    def make_seed(self):
-        return Mnemonic('english').make_seed(num_bits=256, prefix=SEED_PREFIX)
-
     def can_sign_without_server(self):
-        return self.master_private_keys.get('x2/') is not None
+        return not self.keystores['x2/'].is_watching_only()
+
+    def get_user_id(self):
+        return get_user_id(self.storage)
 
     def get_max_amount(self, config, inputs, recipient, fee):
         from electrum.transaction import Transaction
@@ -217,14 +210,14 @@ class Wallet_2fa(Multisig_Wallet):
         for i in inputs:
             self.add_input_info(i)
         xf = self.extra_fee()
+        _type, addr = recipient
         if xf and sendable >= xf:
             billing_address = self.billing_info['billing_address']
             sendable -= xf
-            outputs = [(TYPE_ADDRESS, recipient, sendable),
+            outputs = [(_type, addr, sendable),
                        (TYPE_ADDRESS, billing_address, xf)]
         else:
-            outputs = [(TYPE_ADDRESS, recipient, sendable)]
-
+            outputs = [(_type, addr, sendable)]
         dummy_tx = Transaction.from_io(inputs, outputs)
         if fee is None:
             fee = self.estimate_fee(config, dummy_tx.estimated_size())
@@ -244,7 +237,7 @@ class Wallet_2fa(Multisig_Wallet):
 
     def make_unsigned_transaction(self, coins, outputs, config,
                                   fixed_fee=None, change_addr=None):
-        mk_tx = lambda o: BIP32_Wallet.make_unsigned_transaction(
+        mk_tx = lambda o: Multisig_Wallet.make_unsigned_transaction(
             self, coins, o, config, fixed_fee, change_addr)
         fee = self.extra_fee()
         if fee:
@@ -264,7 +257,7 @@ class Wallet_2fa(Multisig_Wallet):
         return tx
 
     def sign_transaction(self, tx, password):
-        BIP32_Wallet.sign_transaction(self, tx, password)
+        Multisig_Wallet.sign_transaction(self, tx, password)
         if tx.is_complete():
             return
         if not self.auth_code:
@@ -279,27 +272,24 @@ class Wallet_2fa(Multisig_Wallet):
             tx.update(raw_tx)
         self.print_error("twofactor: is complete", tx.is_complete())
 
-    def get_user_id(self):
-        def make_long_id(xpub_hot, xpub_cold):
-            return bitcoin.sha256(''.join(sorted([xpub_hot, xpub_cold])))
-        xpub_hot = self.master_public_keys["x1/"]
-        xpub_cold = self.master_public_keys["x2/"]
-        long_id = make_long_id(xpub_hot, xpub_cold)
-        short_id = hashlib.sha256(long_id).hexdigest()
-        return long_id, short_id
 
 # Utility functions
 
+def get_user_id(storage):
+    def make_long_id(xpub_hot, xpub_cold):
+        return bitcoin.sha256(''.join(sorted([xpub_hot, xpub_cold])))
+    xpub1 = storage.get('x1/')['xpub']
+    xpub2 = storage.get('x2/')['xpub']
+    long_id = make_long_id(xpub1, xpub2)
+    short_id = hashlib.sha256(long_id).hexdigest()
+    return long_id, short_id
+
 def make_xpub(xpub, s):
     _, _, _, c, cK = deserialize_xkey(xpub)
     cK2, c2 = bitcoin._CKD_pub(cK, c, s)
     xpub2 = ("0488B21E" + "00" + "00000000" + "00000000").decode("hex") + c2 + cK2
     return EncodeBase58Check(xpub2)
 
-def restore_third_key(wallet):
-    long_user_id, short_id = wallet.get_user_id()
-    xpub3 = make_xpub(signing_xpub, long_user_id)
-    wallet.add_master_public_key('x3/', xpub3)
 
 def make_billing_address(wallet, num):
     long_id, short_id = wallet.get_user_id()
@@ -324,14 +314,13 @@ class TrustedCoinPlugin(BasePlugin):
     def is_available(self):
         return True
 
-    def set_enabled(self, wallet, enabled):
-        wallet.storage.put('use_' + self.name, enabled)
-
     def is_enabled(self):
         return True
 
     @hook
     def get_additional_fee(self, wallet, tx):
+        if type(wallet) != Wallet_2fa:
+            return
         address = wallet.billing_info['billing_address']
         for _type, addr, amount in tx.outputs():
             if _type == TYPE_ADDRESS and addr == address:
@@ -345,23 +334,62 @@ class TrustedCoinPlugin(BasePlugin):
         wallet.price_per_tx = dict(billing_info['price_per_tx'])
         return True
 
-    def create_extended_seed(self, wallet, window):
-        seed = wallet.make_seed()
-        window.show_and_verify_seed(seed, is_valid=self.is_valid_seed)
+    def make_seed(self):
+        return Mnemonic('english').make_seed(num_bits=128, prefix=SEED_PREFIX)
 
-        password = window.request_password()
-        wallet.storage.put('seed_version', wallet.seed_version)
-        wallet.storage.put('use_encryption', password is not None)
+    @hook
+    def do_clear(self, window):
+        window.wallet.is_billing = False
 
+    def show_disclaimer(self, wizard):
+        wizard.set_icon(':icons/trustedcoin.png')
+        wizard.stack = []
+        wizard.confirm_dialog(title='Disclaimer', message='\n\n'.join(DISCLAIMER), run_next = lambda x: wizard.run('choose_seed'))
+
+    def choose_seed(self, wizard):
+        title = _('Create or restore')
+        message = _('Do you want to create a new seed, or to restore a wallet using an existing seed?')
+        choices = [
+            ('create_seed', _('Create a new seed')),
+            ('restore_wallet', _('I already have a seed')),
+        ]
+        wizard.choice_dialog(title=title, message=message, choices=choices, run_next=wizard.run)
+
+    def create_seed(self, wizard):
+        seed = self.make_seed()
+        f = lambda x: wizard.request_passphrase(seed, x)
+        wizard.show_seed_dialog(run_next=f, seed_text=seed)
+
+    def xkeys_from_seed(self, seed, passphrase):
         words = seed.split()
-        n = len(words)/2
-        wallet.add_xprv_from_seed(' '.join(words[0:n]), 'x1/', password)
-        wallet.add_xpub_from_seed(' '.join(words[n:]), 'x2/')
+        n = len(words)
+        # old version use long seed phrases
+        if n >= 24:
+            assert passphrase == ''
+            xprv1, xpub1 = keystore.xkeys_from_seed(' '.join(words[0:12]), '', "m/")
+            xprv2, xpub2 = keystore.xkeys_from_seed(' '.join(words[12:]), '', "m/")
+        elif n==12:
+            xprv1, xpub1 = keystore.xkeys_from_seed(seed, passphrase, "m/0'/")
+            xprv2, xpub2 = keystore.xkeys_from_seed(seed, passphrase, "m/1'/")
+        else:
+            raise BaseException('unrecognized seed length')
+        return xprv1, xpub1, xprv2, xpub2
 
-        wallet.storage.write()
+    def create_keystore(self, wizard, seed, passphrase):
+        # this overloads the wizard's method
+        xprv1, xpub1, xprv2, xpub2 = self.xkeys_from_seed(seed, passphrase)
+        k1 = keystore.from_xprv(xprv1)
+        k2 = keystore.from_xpub(xpub2)
+        wizard.request_password(run_next=lambda pw: self.on_password(wizard, pw, k1, k2))
 
+    def on_password(self, wizard, password, k1, k2):
+        k1.update_password(None, password)
+        wizard.storage.put('use_encryption', bool(password))
+        wizard.storage.put('x1/', k1.dump())
+        wizard.storage.put('x2/', k2.dump())
+        wizard.storage.write()
         msg = [
-            _("Your wallet file is: %s.")%os.path.abspath(wallet.storage.path),
+            _("Your wallet file is: %s.")%os.path.abspath(wizard.storage.path),
             _("You need to be online in order to complete the creation of "
               "your wallet.  If you generated your seed on an offline "
               'computer, click on "%s" to close this window, move your '
@@ -370,55 +398,82 @@ class TrustedCoinPlugin(BasePlugin):
             _('If you are online, click on "%s" to continue.') % _('Next')
         ]
         msg = '\n\n'.join(msg)
-        self.confirm(window, msg)
+        wizard.stack = []
+        wizard.confirm_dialog(title='', message=msg, run_next = lambda x: wizard.run('create_remote_key'))
 
-    @hook
-    def do_clear(self, window):
-        window.wallet.is_billing = False
+    def restore_wallet(self, wizard):
+        wizard.opt_bip39 = False
+        wizard.opt_ext = True
+        title = _("Restore two-factor Wallet")
+        f = lambda seed, is_bip39, is_ext: wizard.run('on_restore_seed', seed, is_ext)
+        wizard.restore_seed_dialog(run_next=f, test=self.is_valid_seed)
 
-    def on_restore_wallet(self, wallet, wizard):
-        assert isinstance(wallet, self.wallet_class)
+    def on_restore_seed(self, wizard, seed, is_ext):
+        f = lambda x: self.restore_choice(wizard, seed, x)
+        wizard.passphrase_dialog(run_next=f) if is_ext else f('')
 
-        seed = wizard.request_seed(RESTORE_MSG, is_valid=self.is_valid_seed)
-        password = wizard.request_password()
+    def restore_choice(self, wizard, seed, passphrase):
+        wizard.set_icon(':icons/trustedcoin.png')
+        wizard.stack = []
+        title = _('Restore 2FA wallet')
+        msg = ' '.join([
+            'You are going to restore a wallet protected with two-factor authentication.',
+            'Do you want to keep using two-factor authentication with this wallet,',
+            'or do you want to disable it, and have two master private keys in your wallet?'
+        ])
+        choices = [('keep', 'Keep'), ('disable', 'Disable')]
+        f = lambda x: self.on_choice(wizard, seed, passphrase, x)
+        wizard.choice_dialog(choices=choices, message=msg, title=title, run_next=f)
 
-        wallet.add_seed(seed, password)
-        words = seed.split()
-        n = len(words)/2
-        wallet.add_xprv_from_seed(' '.join(words[0:n]), 'x1/', password)
-        wallet.add_xprv_from_seed(' '.join(words[n:]), 'x2/', password)
+    def on_choice(self, wizard, seed, passphrase, x):
+        if x == 'disable':
+            f = lambda pw: wizard.run('on_restore_pw', seed, passphrase, pw)
+            wizard.request_password(run_next=f)
+        else:
+            self.create_keystore(wizard, seed, passphrase)
 
-        restore_third_key(wallet)
-        wallet.create_main_account()
-        return wallet
-
-    def create_remote_key(self, wallet, window):
-        email = self.accept_terms_of_use(window)
-        xpub_hot = wallet.master_public_keys["x1/"]
-        xpub_cold = wallet.master_public_keys["x2/"]
-
-        # Generate third key deterministically.
-        long_user_id, short_id = wallet.get_user_id()
+    def on_restore_pw(self, wizard, seed, passphrase, password):
+        storage = wizard.storage
+        xprv1, xpub1, xprv2, xpub2 = self.xkeys_from_seed(seed, passphrase)
+        k1 = keystore.from_xprv(xprv1)
+        k2 = keystore.from_xprv(xprv2)
+        k1.add_seed(seed)
+        k1.update_password(None, password)
+        k2.update_password(None, password)
+        storage.put('x1/', k1.dump())
+        storage.put('x2/', k2.dump())
+        long_user_id, short_id = get_user_id(storage)
         xpub3 = make_xpub(signing_xpub, long_user_id)
+        k3 = keystore.from_xpub(xpub3)
+        storage.put('use_encryption', bool(password))
+        storage.put('x3/', k3.dump())
+        wizard.wallet = Wallet_2fa(storage)
+        wizard.create_addresses()
 
+    def create_remote_key(self, wizard):
+        email = self.accept_terms_of_use(wizard)
+        xpub1 = wizard.storage.get('x1/')['xpub']
+        xpub2 = wizard.storage.get('x2/')['xpub']
+        # Generate third key deterministically.
+        long_user_id, short_id = get_user_id(wizard.storage)
+        xpub3 = make_xpub(signing_xpub, long_user_id)
         # secret must be sent by the server
         try:
-            r = server.create(xpub_hot, xpub_cold, email)
+            r = server.create(xpub1, xpub2, email)
         except socket.error:
-            window.show_message('Server not reachable, aborting')
+            wizard.show_message('Server not reachable, aborting')
             return
         except TrustedCoinException as e:
             if e.status_code == 409:
                 r = None
             else:
                 raise e
-
         if r is None:
             otp_secret = None
         else:
             otp_secret = r.get('otp_secret')
             if not otp_secret:
-                window.show_message(_('Error'))
+                wizard.show_message(_('Error'))
                 return
             _xpub3 = r['xpubkey_cosigner']
             _id = r['id']
@@ -426,9 +481,72 @@ class TrustedCoinPlugin(BasePlugin):
                 assert _id == short_id, ("user id error", _id, short_id)
                 assert xpub3 == _xpub3, ("xpub3 error", xpub3, _xpub3)
             except Exception as e:
-                window.show_message(str(e))
+                wizard.show_message(str(e))
                 return
+        self.check_otp(wizard, short_id, otp_secret, xpub3)
 
-        if self.setup_google_auth(window, short_id, otp_secret):
-            wallet.add_master_public_key('x3/', xpub3)
-            wallet.create_main_account()
+    def check_otp(self, wizard, short_id, otp_secret, xpub3):
+        otp, reset = self.request_otp_dialog(wizard, short_id, otp_secret)
+        if otp:
+            self.do_auth(wizard, short_id, otp, xpub3)
+        elif reset:
+            wizard.opt_bip39 = False
+            wizard.opt_ext = True
+            f = lambda seed, is_bip39, is_ext: wizard.run('on_reset_seed', short_id, seed, is_ext, xpub3)
+            wizard.restore_seed_dialog(run_next=f, test=self.is_valid_seed)
+
+    def on_reset_seed(self, wizard, short_id, seed, is_ext, xpub3):
+        f = lambda passphrase: wizard.run('on_reset_auth', short_id, seed, passphrase, xpub3)
+        wizard.passphrase_dialog(run_next=f) if is_ext else f('')
+
+    def do_auth(self, wizard, short_id, otp, xpub3):
+        try:
+            server.auth(short_id, otp)
+        except:
+            wizard.show_message(_('Incorrect password'))
+            return
+        k3 = keystore.from_xpub(xpub3)
+        wizard.storage.put('x3/', k3.dump())
+        wizard.storage.put('use_trustedcoin', True)
+        wizard.storage.write()
+        wizard.wallet = Wallet_2fa(wizard.storage)
+        wizard.run('create_addresses')
+
+    def on_reset_auth(self, wizard, short_id, seed, passphrase, xpub3):
+        xprv1, xpub1, xprv2, xpub2 = self.xkeys_from_seed(seed, passphrase)
+        try:
+            assert xpub1 == wizard.storage.get('x1/')['xpub']
+            assert xpub2 == wizard.storage.get('x2/')['xpub']
+        except:
+            wizard.show_message(_('Incorrect seed'))
+            return
+        r = server.get_challenge(short_id)
+        challenge = r.get('challenge')
+        message = 'TRUSTEDCOIN CHALLENGE: ' + challenge
+        def f(xprv):
+            from electrum.bitcoin import deserialize_xkey, bip32_private_key, regenerate_key, is_compressed
+            _, _, _, c, k = deserialize_xkey(xprv)
+            pk = bip32_private_key([0, 0], k, c)
+            key = regenerate_key(pk)
+            compressed = is_compressed(pk)
+            sig = key.sign_message(message, compressed)
+            return base64.b64encode(sig)
+
+        signatures = [f(x) for x in [xprv1, xprv2]]
+        r = server.reset_auth(short_id, challenge, signatures)
+        new_secret = r.get('otp_secret')
+        if not new_secret:
+            wizard.show_message(_('Request rejected by server'))
+            return
+        self.check_otp(wizard, short_id, new_secret, xpub3)
+
+    @hook
+    def get_action(self, storage):
+        if storage.get('wallet_type') != '2fa':
+            return
+        if not storage.get('x1/'):
+            return self, 'show_disclaimer'
+        if not storage.get('x2/'):
+            return self, 'show_disclaimer'
+        if not storage.get('x3/'):
+            return self, 'create_remote_key'
